/* udp-echo.c
 *
 * Similar to the udp echo service on port 7, in case the internal
 * service is disabled. */

/* Copyright (C) 2002 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "echo.h"

#define MAX_PACKET 2000
#define TRACE 0

#ifndef SOL_IP
# define SOL_IP IPPROTO_IP
#endif

int
make_socket(int port)
{
  int s;
  struct sockaddr_in addr;
  
  addr.sin_family = AF_INET;

  /* The port to listen on */
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = INADDR_ANY;
  
  s = socket(PF_INET, SOCK_DGRAM, 0);  /* FIXME: Use PROTO_UDP instead? */
  if (s < 0)
    {
      perror("socket() failed");
      exit(EXIT_FAILURE);
    }

  if (bind(s, (struct sockaddr *) &addr, sizeof(addr)) < 0)
    {
      perror("bind() failed");
      exit(EXIT_FAILURE);
    }

  return s;
}

/* Reads packet, gets ttl from the packet contents, and updates the
 * ttl on the socket accordingly */
static int
read_packet(int s, unsigned *length, unsigned char *packet,
	    struct sockaddr_in *peer)
{
  unsigned packet_length;
  socklen_t peer_length = sizeof(*peer);
  int res;
 
  memset(packet, 0, *length);

  do
    res = recvfrom(s, packet, *length, 0,
		   (struct sockaddr *) peer, &peer_length);
  while (res < 0 && errno == EINTR);

  if (res < 0)
    {
      perror("recv failed");
      return 0;
    }

  packet_length = res;

  if (peer_length != sizeof(*peer))
    {
      fprintf(stderr, "Strange name length, ignoring packet\n");
      return 0;
    }

#if TRACE
  fprintf(stderr, "received packet of length %d\n", packet_length);
#endif
  if (packet_length >= 3
      && packet[0] == ECHO_PROTOCOL_VERSION
      && packet[1]
      && packet[2])
    {
      int ttl = packet[2];
#if TRACE
      fprintf(stderr, "%x,%x,%x\n", packet[0], packet[1], packet[2]);
#endif
      packet_length = packet[1] * ECHO_SIZE_INCREMENT;
      if (packet_length < *length)
	*length = packet_length;

      if (setsockopt(s, SOL_IP, IP_TTL, &ttl, sizeof(ttl)) < 0)
	perror("setsockopt(SOL_IP, IP_TTL) failed");

      return 1;
    }
  else
    return 0;
}

static int
send_packet(int s, unsigned length, unsigned char *packet,
	    struct sockaddr_in *peer)
{
  int res;

  /* Clear ttl field */
  packet[2] = 0;
  
  do
    res = sendto(s, packet, length, 0,
		 (struct sockaddr *) peer, sizeof(*peer));
  while (res < 0 && errno == EINTR);

  if (res < 0)
    {
      perror("send failed");
      return 0;
    }
  return 1;
}

int
main(int argc, char **argv)
{
  int s;
  int port;
  
  argc--; argv++;

  if (!argv[0])
    {
      fprintf(stderr, "Usage: udp-echo PORT\n");
      return EXIT_FAILURE;
    }
  
  port = atoi(argv[0]);
  if (port <= 0)
    {
      fprintf(stderr, "Bad port\n");
	return EXIT_FAILURE;
    }
  
  s = make_socket(port);
  if (s < 0)
    return EXIT_FAILURE;
  
  for (;;)
    {
      unsigned char packet[MAX_PACKET];
      struct sockaddr_in addr;
      unsigned packet_length = sizeof(packet);

      if (read_packet(s, &packet_length, packet, &addr))
	send_packet(s, packet_length, packet, &addr);
    }
}
